@@ -25,7 +25,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
-#include <config.h>
+#include "config.h"
 
 #include "gtkmain.h"
 #include "gtkmarshalers.h"
@@ -36,6 +36,10 @@
 
 #include "gtkalias.h"
 
+static void            gtk_plug_get_property          (GObject     *object,
+						       guint        prop_id,
+						       GValue      *value,
+						       GParamSpec  *pspec);
 static void            gtk_plug_finalize              (GObject          *object);
 static void            gtk_plug_realize               (GtkWidget        *widget);
 static void            gtk_plug_unrealize             (GtkWidget        *widget);
@@ -64,6 +68,12 @@ typedef struct
   GdkModifierType	 accelerator_mods;
 } GrabbedKey;
 
+enum {
+  PROP_0,
+  PROP_EMBEDDED,
+  PROP_SOCKET_WINDOW
+};
+
 enum {
   EMBEDDED,
   LAST_SIGNAL
@@ -73,6 +83,28 @@ static guint plug_signals[LAST_SIGNAL] = { 0 };
 
 G_DEFINE_TYPE (GtkPlug, gtk_plug, GTK_TYPE_WINDOW)
 
+static void
+gtk_plug_get_property (GObject    *object,
+		       guint       prop_id,
+		       GValue     *value,
+		       GParamSpec *pspec)
+{
+  GtkPlug *plug = GTK_PLUG (object);
+
+  switch (prop_id)
+    {
+    case PROP_EMBEDDED:
+      g_value_set_boolean (value, plug->socket_window != NULL);
+      break;
+    case PROP_SOCKET_WINDOW:
+      g_value_set_object (value, plug->socket_window);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
 static void
 gtk_plug_class_init (GtkPlugClass *class)
 {
@@ -83,6 +115,7 @@ gtk_plug_class_init (GtkPlugClass *class)
 
   bin_class = g_type_class_peek (GTK_TYPE_BIN);
 
+  gobject_class->get_property = gtk_plug_get_property;
   gobject_class->finalize = gtk_plug_finalize;
   
   widget_class->realize = gtk_plug_realize;
@@ -104,6 +137,42 @@ gtk_plug_class_init (GtkPlugClass *class)
   window_class->set_focus = gtk_plug_set_focus;
   window_class->keys_changed = gtk_plug_keys_changed;
 
+  /**
+   * GtkPlug:embedded:
+   *
+   * %TRUE if the plug is embedded in a socket.
+   *
+   * Since: 2.12
+   */
+  g_object_class_install_property (gobject_class,
+				   PROP_EMBEDDED,
+				   g_param_spec_boolean ("embedded",
+							 P_("Embedded"),
+							 P_("Whether or not the plug is embedded"),
+							 FALSE,
+							 GTK_PARAM_READABLE));
+
+  /**
+   * GtkPlug:socket-window:
+   *
+   * The window of the socket the plug is embedded in.
+   *
+   * Since: 2.14
+   */
+  g_object_class_install_property (gobject_class,
+				   PROP_SOCKET_WINDOW,
+				   g_param_spec_object ("socket-window",
+							P_("Socket Window"),
+							P_("The window of the socket the plug is embedded in"),
+							GDK_TYPE_WINDOW,
+							GTK_PARAM_READABLE));
+
+  /**
+   * GtkPlug::embedded:
+   * @plug: the object on which the signal was emitted
+   *
+   * Gets emitted when the plug becomes embedded in a socket.
+   */ 
   plug_signals[EMBEDDED] =
     g_signal_new (I_("embedded"),
 		  G_OBJECT_CLASS_TYPE (class),
@@ -150,7 +219,7 @@ gtk_plug_set_is_child (GtkPlug  *plug,
       if (GTK_WIDGET_MAPPED (plug))
 	gtk_widget_unmap (GTK_WIDGET (plug));
       
-      GTK_WIDGET_UNSET_FLAGS (plug, GTK_TOPLEVEL);
+      _gtk_window_set_is_toplevel (GTK_WINDOW (plug), FALSE);
       gtk_container_set_resize_mode (GTK_CONTAINER (plug), GTK_RESIZE_PARENT);
 
       _gtk_widget_propagate_hierarchy_changed (GTK_WIDGET (plug), GTK_WIDGET (plug));
@@ -165,7 +234,7 @@ gtk_plug_set_is_child (GtkPlug  *plug,
       plug->modality_group = gtk_window_group_new ();
       gtk_window_group_add_window (plug->modality_group, GTK_WINDOW (plug));
       
-      GTK_WIDGET_SET_FLAGS (plug, GTK_TOPLEVEL);
+      _gtk_window_set_is_toplevel (GTK_WINDOW (plug), TRUE);
       gtk_container_set_resize_mode (GTK_CONTAINER (plug), GTK_RESIZE_QUEUE);
 
       _gtk_widget_propagate_hierarchy_changed (GTK_WIDGET (plug), NULL);
@@ -193,6 +262,42 @@ gtk_plug_get_id (GtkPlug *plug)
   return _gtk_plug_windowing_get_id (plug);
 }
 
+/**
+ * gtk_plug_get_embedded:
+ * @plug: a #GtkPlug
+ *
+ * Determines whether the plug is embedded in a socket.
+ *
+ * Return value: %TRUE if the plug is embedded in a socket
+ *
+ * Since: 2.14
+ **/
+gboolean
+gtk_plug_get_embedded (GtkPlug *plug)
+{
+  g_return_val_if_fail (GTK_IS_PLUG (plug), FALSE);
+
+  return plug->socket_window != NULL;
+}
+
+/**
+ * gtk_plug_get_socket_window:
+ * @plug: a #GtkPlug
+ *
+ * Retrieves the socket the plug is embedded in.
+ *
+ * Return value: the window of the socket, or %NULL
+ *
+ * Since: 2.14
+ **/
+GdkWindow *
+gtk_plug_get_socket_window (GtkPlug *plug)
+{
+  g_return_val_if_fail (GTK_IS_PLUG (plug), NULL);
+
+  return plug->socket_window;
+}
+
 /**
  * _gtk_plug_add_to_socket:
  * @plug: a #GtkPlug
@@ -219,6 +324,9 @@ _gtk_plug_add_to_socket (GtkPlug   *plug,
   socket_->plug_widget = widget;
 
   plug->socket_window = GTK_WIDGET (socket_)->window;
+  g_object_ref (plug->socket_window);
+  g_signal_emit (plug, plug_signals[EMBEDDED], 0);
+  g_object_notify (G_OBJECT (plug), "embedded");
 
   if (GTK_WIDGET_REALIZED (widget))
     {
@@ -228,7 +336,7 @@ _gtk_plug_add_to_socket (GtkPlug   *plug,
 
   gtk_widget_set_parent (widget, GTK_WIDGET (socket_));
 
-  g_signal_emit_by_name (socket_, "plug_added");
+  g_signal_emit_by_name (socket_, "plug-added");
 }
 
 /**
@@ -243,16 +351,16 @@ void
 _gtk_plug_send_delete_event (GtkWidget *widget)
 {
   GdkEvent *event = gdk_event_new (GDK_DELETE);
-  
+
   event->any.window = g_object_ref (widget->window);
   event->any.send_event = FALSE;
 
-  gtk_widget_ref (widget);
-  
+  g_object_ref (widget);
+
   if (!gtk_widget_event (widget, event))
     gtk_widget_destroy (widget);
-  
-  gtk_widget_unref (widget);
+
+  g_object_unref (widget);
 
   gdk_event_free (event);
 }
@@ -304,11 +412,14 @@ _gtk_plug_remove_from_socket (GtkPlug   *plug,
   socket_->same_app = FALSE;
 
   plug->same_app = FALSE;
-  plug->socket_window = NULL;
-
+  if (plug->socket_window != NULL)
+    {
+      g_object_unref (plug->socket_window);
+      plug->socket_window = NULL;
+    }
   gtk_plug_set_is_child (plug, FALSE);
-		    
-  g_signal_emit_by_name (socket_, "plug_removed", &result);
+
+  g_signal_emit_by_name (socket_, "plug-removed", &result);
   if (!result)
     gtk_widget_destroy (GTK_WIDGET (socket_));
 
@@ -361,25 +472,31 @@ gtk_plug_construct_for_display (GtkPlug         *plug,
       gpointer user_data = NULL;
 
       plug->socket_window = gdk_window_lookup_for_display (display, socket_id);
-      
       if (plug->socket_window)
-	gdk_window_get_user_data (plug->socket_window, &user_data);
+	{
+	  gdk_window_get_user_data (plug->socket_window, &user_data);
+
+	  if (user_data)
+	    {
+	      if (GTK_IS_SOCKET (user_data))
+		_gtk_plug_add_to_socket (plug, user_data);
+	      else
+		{
+		  g_warning (G_STRLOC "Can't create GtkPlug as child of non-GtkSocket");
+		  plug->socket_window = NULL;
+		}
+	    }
+	  else
+	    g_object_ref (plug->socket_window);
+	}
       else
 	plug->socket_window = gdk_window_foreign_new_for_display (display, socket_id);
-	  
-      if (user_data)
-	{
-	  if (GTK_IS_SOCKET (user_data))
-	    _gtk_plug_add_to_socket (plug, user_data);
-	  else
-	    {
-	      g_warning (G_STRLOC "Can't create GtkPlug as child of non-GtkSocket");
-	      plug->socket_window = NULL;
-	    }
-	}
 
-      if (plug->socket_window)
+      if (plug->socket_window) {
 	g_signal_emit (plug, plug_signals[EMBEDDED], 0);
+
+        g_object_notify (G_OBJECT (plug), "embedded");
+      }
     }
 }
 
@@ -438,17 +555,15 @@ gtk_plug_finalize (GObject *object)
 static void
 gtk_plug_unrealize (GtkWidget *widget)
 {
-  GtkPlug *plug;
-
-  g_return_if_fail (GTK_IS_PLUG (widget));
-
-  plug = GTK_PLUG (widget);
+  GtkPlug *plug = GTK_PLUG (widget);
 
   if (plug->socket_window != NULL)
     {
       gdk_window_set_user_data (plug->socket_window, NULL);
       g_object_unref (plug->socket_window);
       plug->socket_window = NULL;
+
+      g_object_notify (G_OBJECT (widget), "embedded");
     }
 
   if (!plug->same_app)
@@ -459,24 +574,19 @@ gtk_plug_unrealize (GtkWidget *widget)
       gtk_window_group_remove_window (plug->modality_group, GTK_WINDOW (plug));
       g_object_unref (plug->modality_group);
     }
-  
-  if (GTK_WIDGET_CLASS (gtk_plug_parent_class)->unrealize)
-    (* GTK_WIDGET_CLASS (gtk_plug_parent_class)->unrealize) (widget);
+
+  GTK_WIDGET_CLASS (gtk_plug_parent_class)->unrealize (widget);
 }
 
 static void
 gtk_plug_realize (GtkWidget *widget)
 {
-  GtkWindow *window;
-  GtkPlug *plug;
+  GtkWindow *window = GTK_WINDOW (widget);
+  GtkPlug *plug = GTK_PLUG (widget);
   GdkWindowAttr attributes;
   gint attributes_mask;
 
-  g_return_if_fail (GTK_IS_PLUG (widget));
-
   GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
-  window = GTK_WINDOW (widget);
-  plug = GTK_PLUG (widget);
 
   attributes.window_type = GDK_WINDOW_CHILD;	/* XXX GDK_WINDOW_PLUG ? */
   attributes.title = window->title;
