@@ -38,6 +38,8 @@
 #include "gtkbindings.h"
 #include "gtkprivate.h"
 #include "gtkintl.h"
+#include "gtkbuildable.h"
+#include "gtkbuilderprivate.h"
 #include "gtkalias.h"
 
 
@@ -56,7 +58,7 @@ typedef struct _GtkScaleMark GtkScaleMark;
 struct _GtkScaleMark
 {
   gdouble          value;
-  const gchar     *markup;
+  gchar           *markup;
   GtkPositionType  position;
 };
 
@@ -110,8 +112,24 @@ static gboolean gtk_scale_expose                  (GtkWidget      *widget,
 static void     gtk_scale_real_get_layout_offsets (GtkScale       *scale,
                                                    gint           *x,
                                                    gint           *y);
+static void     gtk_scale_buildable_interface_init   (GtkBuildableIface *iface);
+static gboolean gtk_scale_buildable_custom_tag_start (GtkBuildable  *buildable,
+                                                      GtkBuilder    *builder,
+                                                      GObject       *child,
+                                                      const gchar   *tagname,
+                                                      GMarkupParser *parser,
+                                                      gpointer      *data);
+static void     gtk_scale_buildable_custom_finished  (GtkBuildable  *buildable,
+                                                      GtkBuilder    *builder,
+                                                      GObject       *child,
+                                                      const gchar   *tagname,
+                                                      gpointer       user_data);
+
+
+G_DEFINE_ABSTRACT_TYPE_WITH_CODE (GtkScale, gtk_scale, GTK_TYPE_RANGE,
+                                  G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+                                                         gtk_scale_buildable_interface_init))
 
-G_DEFINE_ABSTRACT_TYPE (GtkScale, gtk_scale, GTK_TYPE_RANGE)
 
 static gboolean
 single_string_accumulator (GSignalInvocationHint *ihint,
@@ -907,6 +925,27 @@ gtk_scale_size_request (GtkWidget      *widget,
     }
 }
 
+static gint
+find_next_pos (GtkWidget      *widget,
+               GSList          *list,
+               gint            *marks,
+               GtkPositionType  pos,
+               gint             match)
+{
+  GSList *m;
+  gint i;
+
+  for (m = list->next, i = 1; m; m = m->next, i++)
+    {
+      GtkScaleMark *mark = m->data;
+
+      if (match == (mark->position == pos))
+        return marks[i];
+    }
+
+  return widget->allocation.width;
+}
+
 static gboolean
 gtk_scale_expose (GtkWidget      *widget,
                   GdkEventExpose *event)
@@ -920,6 +959,7 @@ gtk_scale_expose (GtkWidget      *widget,
   gint focus_padding;
   gint slider_width;
   gint value_spacing;
+  gint min_sep = 4;
 
   gtk_widget_style_get (widget,
                         "focus-padding", &focus_padding,
@@ -943,10 +983,16 @@ gtk_scale_expose (GtkWidget      *widget,
       PangoLayout *layout;
       PangoRectangle logical_rect;
       GSList *m;
+      gint min_pos_before, min_pos_after;
+      gint min_pos, max_pos;
 
       n_marks = _gtk_range_get_stop_positions (range, &marks);
       layout = gtk_widget_create_pango_layout (widget, NULL);
 
+      if (range->orientation == GTK_ORIENTATION_HORIZONTAL)
+        min_pos_before = min_pos_after = widget->allocation.x;
+      else
+        min_pos_before = min_pos_after = widget->allocation.y;
       for (m = priv->marks, i = 0; m; m = m->next, i++)
         {
           GtkScaleMark *mark = m->data;
@@ -958,11 +1004,15 @@ gtk_scale_expose (GtkWidget      *widget,
                 {
                   y1 = widget->allocation.y + range->range_rect.y;
                   y2 = y1 - slider_width / 2;
+                  min_pos = min_pos_before;
+                  max_pos = widget->allocation.x + find_next_pos (widget, m, marks + i, GTK_POS_TOP, 1) - min_sep;
                 }
               else
                 {
                   y1 = widget->allocation.y + range->range_rect.y + range->range_rect.height;
                   y2 = y1 + slider_width / 2;
+                  min_pos = min_pos_after;
+                  max_pos = widget->allocation.x + find_next_pos (widget, m, marks + i, GTK_POS_TOP, 0) - min_sep;
                 }
 
               gtk_paint_vline (widget->style, widget->window, state_type,
@@ -972,15 +1022,27 @@ gtk_scale_expose (GtkWidget      *widget,
                 {
                   pango_layout_set_markup (layout, mark->markup, -1);
                   pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
-              
+
                   x3 = x1 - logical_rect.width / 2;
+                  if (x3 < min_pos)
+                    x3 = min_pos;
+                  if (x3 + logical_rect.width > max_pos)
+                        x3 = max_pos - logical_rect.width;
+                  if (x3 < widget->allocation.x)
+                     x3 = widget->allocation.x;
                   if (mark->position == GTK_POS_TOP)
-                    y3 = y2 - value_spacing - logical_rect.height;
+                    {
+                      y3 = y2 - value_spacing - logical_rect.height;
+                      min_pos_before = x3 + logical_rect.width + min_sep;
+                    }
                   else
-                    y3 = y2 + value_spacing; 
+                    {
+                      y3 = y2 + value_spacing;
+                      min_pos_after = x3 + logical_rect.width + min_sep;
+                    }
 
                   gtk_paint_layout (widget->style, widget->window, state_type,
-                                    FALSE, NULL, widget, "scale-mark", 
+                                    FALSE, NULL, widget, "scale-mark",
                                     x3, y3, layout);
                 }
             }
@@ -990,11 +1052,15 @@ gtk_scale_expose (GtkWidget      *widget,
                 {
                   x1 = widget->allocation.x + range->range_rect.x;
                   x2 = widget->allocation.x + range->range_rect.x - slider_width / 2;
+                  min_pos = min_pos_before;
+                  max_pos = widget->allocation.y + find_next_pos (widget, m, marks + i, GTK_POS_LEFT, 1) - min_sep;
                 }
               else
                 {
                   x1 = widget->allocation.x + range->range_rect.x + range->range_rect.width;
                   x2 = widget->allocation.x + range->range_rect.x + range->range_rect.width + slider_width / 2;
+                  min_pos = min_pos_after;
+                  max_pos = widget->allocation.y + find_next_pos (widget, m, marks + i, GTK_POS_LEFT, 0) - min_sep;
                 }
               y1 = widget->allocation.y + marks[i];
 
@@ -1006,14 +1072,26 @@ gtk_scale_expose (GtkWidget      *widget,
                   pango_layout_set_markup (layout, mark->markup, -1);
                   pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
               
-                  if (mark->position == GTK_POS_LEFT)
-                    x3 = x2 - value_spacing - logical_rect.width;
-                  else
-                    x3 = x2 + value_spacing; 
                   y3 = y1 - logical_rect.height / 2;
+                  if (y3 < min_pos)
+                    y3 = min_pos;
+                  if (y3 + logical_rect.height > max_pos)
+                    y3 = max_pos - logical_rect.height;
+                  if (y3 < widget->allocation.y)
+                    y3 = widget->allocation.y;
+                  if (mark->position == GTK_POS_LEFT)
+                    {
+                      x3 = x2 - value_spacing - logical_rect.width;
+                      min_pos_before = y3 + logical_rect.height + min_sep;
+                    }
+                  else
+                    {
+                      x3 = x2 + value_spacing;
+                      min_pos_after = y3 + logical_rect.height + min_sep;
+                    }
 
                   gtk_paint_layout (widget->style, widget->window, state_type,
-                                    FALSE, NULL, widget, "scale-mark", 
+                                    FALSE, NULL, widget, "scale-mark",
                                     x3, y3, layout);
                 }
             }
@@ -1293,6 +1371,18 @@ gtk_scale_clear_marks (GtkScale *scale)
   priv->marks = NULL;
 
   _gtk_range_set_stop_values (GTK_RANGE (scale), NULL, 0);
+
+  gtk_widget_queue_resize (GTK_WIDGET (scale));
+}
+
+static gint
+compare_marks (gpointer a, gpointer b)
+{
+  GtkScaleMark *ma, *mb;
+
+  ma = a; mb = b;
+
+  return (gint) (ma->value - mb->value);
 }
 
 /**
@@ -1336,7 +1426,8 @@ gtk_scale_add_mark (GtkScale        *scale,
   mark->markup = g_strdup (markup);
   mark->position = position;
  
-  priv->marks = g_slist_prepend (priv->marks, mark);
+  priv->marks = g_slist_insert_sorted (priv->marks, mark,
+                                       (GCompareFunc) compare_marks);
 
   n = g_slist_length (priv->marks);
   values = g_new (gdouble, n);
@@ -1349,8 +1440,243 @@ gtk_scale_add_mark (GtkScale        *scale,
   _gtk_range_set_stop_values (GTK_RANGE (scale), values, n);
 
   g_free (values);
+
+  gtk_widget_queue_resize (GTK_WIDGET (scale));
 }
 
+static GtkBuildableIface *parent_buildable_iface;
+
+static void
+gtk_scale_buildable_interface_init (GtkBuildableIface *iface)
+{
+  parent_buildable_iface = g_type_interface_peek_parent (iface);
+  iface->custom_tag_start = gtk_scale_buildable_custom_tag_start;
+  iface->custom_finished = gtk_scale_buildable_custom_finished;
+}
+
+typedef struct
+{
+  GtkScale *scale;
+  GtkBuilder *builder;
+  GSList *marks;
+} MarksSubparserData;
+
+typedef struct
+{
+  gdouble value;
+  GtkPositionType position;
+  GString *markup;
+  gchar *context;
+  gboolean translatable;
+} MarkData;
+
+static void
+mark_data_free (MarkData *data)
+{
+  g_string_free (data->markup, TRUE);
+  g_free (data->context);
+  g_slice_free (MarkData, data);
+}
+
+static void
+marks_start_element (GMarkupParseContext *context,
+                     const gchar         *element_name,
+                     const gchar        **names,
+                     const gchar        **values,
+                     gpointer             user_data,
+                     GError             **error)
+{
+  MarksSubparserData *parser_data = (MarksSubparserData*)user_data;
+  guint i;
+  gint line_number, char_number;
+
+  if (strcmp (element_name, "marks") == 0)
+   ;
+  else if (strcmp (element_name, "mark") == 0)
+    {
+      gdouble value = 0;
+      gboolean has_value = FALSE;
+      GtkPositionType position = GTK_POS_BOTTOM;
+      const gchar *msg_context = NULL;
+      gboolean translatable = FALSE;
+      MarkData *mark;
+
+      for (i = 0; names[i]; i++)
+        {
+          if (strcmp (names[i], "translatable") == 0)
+            {
+              if (!_gtk_builder_boolean_from_string (values[i], &translatable, error))
+                return;
+            }
+          else if (strcmp (names[i], "comments") == 0)
+            {
+              /* do nothing, comments are for translators */
+            }
+          else if (strcmp (names[i], "context") == 0)
+            msg_context = values[i];
+          else if (strcmp (names[i], "value") == 0)
+            {
+              GValue gvalue = { 0, };
+
+              if (!gtk_builder_value_from_string_type (parser_data->builder, G_TYPE_DOUBLE, values[i], &gvalue, error))
+                return;
+
+              value = g_value_get_double (&gvalue);
+              has_value = TRUE;
+            }
+          else if (strcmp (names[i], "position") == 0)
+            {
+              GValue gvalue = { 0, };
+
+              if (!gtk_builder_value_from_string_type (parser_data->builder, GTK_TYPE_POSITION_TYPE, values[i], &gvalue, error))
+                return;
+
+              position = g_value_get_enum (&gvalue);
+            }
+          else
+            {
+              g_markup_parse_context_get_position (context,
+                                                   &line_number,
+                                                   &char_number);
+              g_set_error (error,
+                           GTK_BUILDER_ERROR,
+                           GTK_BUILDER_ERROR_INVALID_ATTRIBUTE,
+                           "%s:%d:%d '%s' is not a valid attribute of <%s>",
+                           "<input>",
+                           line_number, char_number, names[i], "mark");
+              return;
+            }
+        }
+
+      if (!has_value)
+        {
+          g_markup_parse_context_get_position (context,
+                                               &line_number,
+                                               &char_number);
+          g_set_error (error,
+                       GTK_BUILDER_ERROR,
+                       GTK_BUILDER_ERROR_MISSING_ATTRIBUTE,
+                       "%s:%d:%d <%s> requires attribute \"%s\"",
+                       "<input>",
+                       line_number, char_number, "mark",
+                       "value");
+          return;
+        }
+
+      mark = g_slice_new (MarkData);
+      mark->value = value;
+      mark->position = position;
+      mark->markup = g_string_new ("");
+      mark->context = g_strdup (msg_context);
+      mark->translatable = translatable;
+
+      parser_data->marks = g_slist_prepend (parser_data->marks, mark);
+    }
+  else
+    {
+      g_markup_parse_context_get_position (context,
+                                           &line_number,
+                                           &char_number);
+      g_set_error (error,
+                   GTK_BUILDER_ERROR,
+                   GTK_BUILDER_ERROR_MISSING_ATTRIBUTE,
+                   "%s:%d:%d unsupported tag for GtkScale: \"%s\"",
+                   "<input>",
+                   line_number, char_number, element_name);
+      return;
+    }
+}
+
+static void
+marks_text (GMarkupParseContext  *context,
+            const gchar          *text,
+            gsize                 text_len,
+            gpointer              user_data,
+            GError              **error)
+{
+  MarksSubparserData *data = (MarksSubparserData*)user_data;
+
+  if (strcmp (g_markup_parse_context_get_element (context), "mark") == 0)
+    {
+      MarkData *mark = data->marks->data;
+
+      g_string_append_len (mark->markup, text, text_len);
+    }
+}
+
+static const GMarkupParser marks_parser =
+  {
+    marks_start_element,
+    NULL,
+    marks_text,
+  };
+
+
+static gboolean
+gtk_scale_buildable_custom_tag_start (GtkBuildable  *buildable,
+                                      GtkBuilder    *builder,
+                                      GObject       *child,
+                                      const gchar   *tagname,
+                                      GMarkupParser *parser,
+                                      gpointer      *data)
+{
+  MarksSubparserData *parser_data;
+
+  if (child)
+    return FALSE;
+
+  if (strcmp (tagname, "marks") == 0)
+    {
+      parser_data = g_slice_new0 (MarksSubparserData);
+      parser_data->scale = GTK_SCALE (buildable);
+      parser_data->marks = NULL;
+
+      *parser = marks_parser;
+      *data = parser_data;
+      return TRUE;
+    }
+
+  return parent_buildable_iface->custom_tag_start (buildable, builder, child,
+                                                   tagname, parser, data);
+}
+
+static void
+gtk_scale_buildable_custom_finished (GtkBuildable *buildable,
+                                     GtkBuilder   *builder,
+                                     GObject      *child,
+                                     const gchar  *tagname,
+                                     gpointer      user_data)
+{
+  GtkScale *scale = GTK_SCALE (buildable);
+  MarksSubparserData *marks_data;
+
+  if (strcmp (tagname, "marks") == 0)
+    {
+      GSList *m;
+      gchar *markup;
+
+      marks_data = (MarksSubparserData *)user_data;
+
+      for (m = marks_data->marks; m; m = m->next)
+        {
+          MarkData *mdata = m->data;
+
+          if (mdata->translatable && mdata->markup->len)
+            markup = _gtk_builder_parser_translate (gtk_builder_get_translation_domain (builder),
+                                                    mdata->context,
+                                                    mdata->markup->str);
+          else
+            markup = mdata->markup->str;
+
+          gtk_scale_add_mark (scale, mdata->value, mdata->position, markup);
+
+          mark_data_free (mdata);
+        }
+
+      g_slist_free (marks_data->marks);
+      g_slice_free (MarksSubparserData, marks_data);
+    }
+}
 
 #define __GTK_SCALE_C__
 #include "gtkaliasdef.c"
